;ROTDISP.ASM       

;Rotator Display controller
;Interrupt Timing routine runs continuously, multiplexing display

    LIST P=16F870

    INCLUDE "P16F870.INC"
    __config    0x3D71  ;Mem prot off, LV Prog OFF, XT Osc, PUT On  WDT Off

	errorlevel 2

    
    #define     SWStrb      PORTA, 5    ;Switch matrix Strobe
    #define     SEGMENTS    PORTB
    #define     R_DMND      PORTC, 4    ;Connected to bidirectional
    #define     L_DMND      PORTC, 5    ;  LED interface.
    #define     SpareD      PORTA, 4

    #define     COMMONANODE Flags, 0
    #define     LONGCOMMAND Flags, 1    ;Set after [cr] has been received
    #define     NEWCHAR     Flags, 2
    #define     LIMIT       Flags, 3
    #define     MOVESLOW    Flags, 4
    #define     NEGFLAG     Flags, 5    ;Value in A is negative
  
 GO_RIGHT   =   1       ;Button values in SWStates
 GO_LEFT    =   2
 GO_SLOW    =   4

 SlowConst  =   d'16'   ;Duty cycle = SlowConst/64 at 0.26s intervals

    MAX232 = 0      ;0 = direct RS232 drive polarity, 1 = inverter used


    KVALUE = d'2048'    ;Vtop / Vref * 1024
    MVALUE = d'15897'   ;Gradient of P (0 - 4095) vs. Theta (1/10th's of Deg)
    MSIGN  = 1          ;Sign of gradient (1 = -Ve, 0 = +Ve)
    CVALUE = d'3758'    ;Y intercept (units of tenths of degree
    ZVALUE = d'1780'

      ;Calibrated values 5 August... M = -15897, C = 3758, Z = 1695
	
    org     0x2100 
Kconst      de      HIGH(KVALUE), LOW(KVALUE)       ;K  Vref multiplier
Mconst      de      HIGH(MVALUE), LOW(MVALUE)       ;M  4096 * 10 * gradient
Msign       de      MSIGN                           ;>= 1 = -ve gradient
Cconst      de      HIGH(CVALUE), LOW(CVALUE)       ;C  offset
Zconst      de      HIGH(ZVALUE), LOW(ZVALUE)       ;K  bearing zero
                                                      ;Units of 0.1 deg
    org     0
;=====================================
    goto    StartUp
;-----------------------
    org 4
;Interrupt routine   - Timer runs continuously counting units of

    movwf   Wreg        ;Save W
    swapf   STATUS , W      ; nb.  movwf does not affect any STATUS bits
    movwf   StsReg      ;Stave STATUS, without affecting it
    movf    FSR, W
    movwf   FSRReg

    btfss   PIR1, RCIF  ;First test for serial character received
    goto    TimerInt    ;IF not, must have been a timer interrupt

    movf    RCREG, W
    movwf   RxChar
    bsf     NEWCHAR        ;Set flag if new serial has arrived

    movlw   "#"             ;Long commands preceded by #, reset stack
    subwf   RxChar, W
    btfss   STATUS, Z
    goto    CharToStack
    movlw   0xA0
    movwf   StackPointer
 
CharToStack                 ;Push all chars onto stack, a [cr] terminates
    movf    StackPointer, W ; the command and resets stack pointer
    movwf   FSR
    movf    RxChar, W
    movwf   INDF
    incf    StackPointer
    btfsc   StackPointer, 6 ;Detect overflow from 0xBF to 0xC0, in which
    decf    StackPointer    ; case freeze at 0xBF for over-long command

    movlw   0x0D        ;Check for [cr] to terminate a command
    subwf   RxChar, W
    btfsc   STATUS, Z
    bsf     LONGCOMMAND     ;Set flag if new long command has arrived

    btfss   RCSTA, OERR     ;Check for any overrun error
    goto    SerialOK
    bcf     RCSTA, CREN     ;In which case, reset the UART receiver
    nop
    bsf     RCSTA, CREN

SerialOK
      ;........BUT......... By now, a timer interrupt may have happened
                            ;  so test for it and execute if necessary
    btfss   INTCON, T0IF    ; 
    goto    OutInt

TimerInt                    ;256 clocks  of Fc/16 = 4.096ms
    bcf     INTCON , T0IF   ; to multiplex the display.
    movlw   0xFF            
	btfss	COMMONANODE	
	movlw	0x00		;All segments off.  Stops LED segments pulling
	movwf	PORTB		;  switch lines, whatever the polarity of display

		;CA display needs +Ve digit drive and inverted segment data
		;CC display neds inverted digit drive and normal segment data

			;Read switch states
	clrf	DigitPatt
	call	SetDigit	;Force all digit drives to off state
	bsf     STATUS, RP0     
    movlw   b'11111111'   
    movwf   PORTB       ;Port B as input, pull ups ensure these default to 1
    bcf     STATUS, RP0
    bcf     SWStrb      ;Enable switch matrix
    goto    $+1     ;Short delay to settle
	goto	$+1
	goto	$+1
	comf	PORTB, W	;Read Switch states (into normal logic)
    movwf   SWStates
    bsf     SWStrb

	bsf     STATUS, RP0
    clrf    PORTB       ;Port B as output again
    bcf     STATUS, RP0

			;Next multiplexed digit
    clrf    DigitPatt

    movlw   0x03
    andwf   IntCounter, W
    sublw   0
    btfsc   STATUS, Z
    bsf     DigitPatt, 0

    movlw   0x03
    andwf   IntCounter, W
    sublw   1
    btfsc   STATUS, Z
    bsf     DigitPatt, 1

    movlw   0x03
    andwf   IntCounter, W
    sublw   2
    btfsc   STATUS, Z
    bsf     DigitPatt, 2

    movlw   0x03
    andwf   IntCounter, W
    sublw   3
    btfsc   STATUS, Z
    bsf     DigitPatt, 3

    movf    IntCounter, W
    andlw   3
    addlw   Digit0      ;Point to associated data

    movwf   FSR         ;Indirect addressing to get the digit
	movf	INDF, W
	btfsc	COMMONANODE	;Invert the segment data for CA displays
	xorlw	0xFF
	movwf	PORTB		;Set segment drive lines
	call	SetDigit	;Place the digit data on drive lines.  Display is now ON
    incf    IntCounter  ;Counts 0 - 255 repratedly

OutInt
    movf    FSRReg, W
    movwf   FSR
    swapf   StsReg , W      ;Restore STATUS
    movwf   STATUS
    swapf   Wreg        ;Restore W
    swapf   Wreg , W    ; nb.  movf affects Z bit

    retfie  
;--------------------------------------
SetDigit	;Single 1 set in DigitPatt B0-B3 for digit drive, or 0 for blank  13 clocks total

	movf	PORTC, W	;Get and keep current state of drive lines
	andlw	0xF0		;  and mask out old digit drive data
	movwf	Ccache	
			
	movf	DigitPatt, W
	btfss	COMMONANODE	;Polarity is correct for CA display
	xorlw	0xFF		;Invert polarity for CC display

	andlw	0x0F		;Mask only to digit drives
	iorwf	Ccache, W	;Mix in saved drive data
	movwf	PORTC

	return
;===================================
MainLoop            ;Return to here from StartUp

    bcf     R_DMND
    bcf     L_DMND
    
    movf    SWStates, W
    andlw   b'00000111'     ;Mask to three buttone only
    btfss   STATUS, Z
    goto    SwitchPressed   ;If no switch activated, do nothing

    goto    NoSwitchPressed
;......................
SwitchPressed
    movlw   GO_SLOW
    subwf   SWStates, W
    btfsc   STATUS, Z
    goto    NoSwitchPressed ;Nothing if Slow button only

    movlw   GO_RIGHT        ;R Button Only
    subwf   SWStates, W
    btfss   STATUS, Z
    goto    NotFastRight

    bsf     R_DMND
    call    CheckLimits
    movlw   d'100'          ;
    call    Delay2Nms       ; to stabilise
    goto    DoneSwitches

NotFastRight
    movlw   GO_LEFT         ;L Button Only
    subwf   SWStates, W
    btfss   STATUS, Z
    goto    NotFastLeft

    bsf     L_DMND
    call    CheckLimits

    movlw   d'100'          ;
    call    Delay2Nms       ;
    goto    DoneSwitches

NotFastLeft
    movlw   GO_RIGHT + GO_SLOW  
    subwf   SWStates, W
    btfss   STATUS, Z
    goto    NotSlowRight

    movf    IntCounter, W   ;Pulse only when IntCounter MOD 64 < SlowConst
    andlw   0x3F
    sublw   SlowConst       ;W = SlowConst - (IntCounter MOD 64)
    btfsc   STATUS, C       ;Valid only if result is +ve, C=1
    bsf     R_DMND          ;Don't perform limit checking on pulsed operation

    btfsc   R_DMND
    goto    NoSwitchPressed ;Only update reading if Motor has pulsed OFF
    goto    DoneSwitches
;,,,,,,,,,
NotSlowRight
    movlw   GO_LEFT + GO_SLOW
    subwf   SWStates, W
    btfss   STATUS, Z
    goto    NotSlowLeft

    movf    IntCounter, W   ;Pulse only when IntCounter MOD 64 < SlowConst
    andlw   0x3F
    sublw   SlowConst       ;W = SlowConst - (IntCounter MOD 64)
    btfsc   STATUS, C       ;Valid only if result is +ve, C=1
    bsf     L_DMND

    btfss   L_DMND
    goto    DoneSwitches    ;Only update reading if Motor has pulsed OFF
    goto    NoSwitchPressed
;........
NotSlowLeft

DoneSwitches
    call    ReadP           ;Rotor position in P0/1
    call    CalcBearing     ;Converts P to Bearing (0-3600 in 0.1deg) in A
    call    BinToBCD        ;Binary value in A to BCD variables

    call    ShowBCD4
    movlw   b'01100011'     ;Generate a degree symbol
    movwf   Digit0          ;Place in LSD

    movlw   'b'
    call    Send232

    swapf   BCDMe, W		;Need to send DP so can't use the SendBCD sub.
    andlw   0x0F
    addlw   0x30
    call    Send232

    movf    BCDMe, W
    andlw   0x0F
    addlw   0x30
    call    Send232

    swapf   BCDLo, W
    andlw   0x0F
    addlw   0x30
    call    Send232

    movlw   "."
    call    Send232

    movf    BCDLo, W
    andlw   0x0F
    addlw   0x30
    call    Send232

;    call    SendBCD         ;Send #dddd on RS232

    movlw   ","
    call    Send232
    movlw   "p"
    call    Send232

    movf    P0, W
    movwf   A0
    movf    P1, W
    movwf   A1
    call    BinToBCD
    call    SendBCD
    call    CrLf

    movlw   d'20'           ;Slow things down a bit and allow interrupts
    call    Delay2Nms       ; to stabilise

NoSwitchPressed
       ;~~~~~~~~~~~~~ Now deal with RS232 commands ~~~~~~~~~~~~

    btfss   NEWCHAR         ;Flag is set from inside interrupt routine
    goto    MainLoop

    bcf     INTCON, T0IE    ;Disable multiplex to prevent corruption of FSR
                            ;  blank display when handling commands
    clrf    DigitPatt
    call    SetDigit
    bcf     NEWCHAR

    movlw   "<"             ;First check for 'Quick' commands which
    subwf   RxChar,W        ;  appear in RxChar - specif characters only
    btfss   STATUS, Z       ;"<" rotate left for 0.5s 
    goto    Not1sLeft

    bsf     L_DMND
    movlw   d'250'
    call    Delay2Nms
    bcf     L_DMND
    call    CheckLimits
    goto    DoneCommand

Not1sLeft
    movlw   ">"             ;
    subwf   RxChar, W          ;">" rotate right for 0.5s
    btfss   STATUS, Z
    goto    Not1sRight

    bsf     R_DMND
    movlw   d'250'
    call    Delay2Nms
    bcf     R_DMND
    call    CheckLimits
    goto    DoneCommand

Not1sRight
    movlw   ","          
    subwf   RxChar, W          ;"<" rotate left for 0.1s
    btfss   STATUS, Z
    goto    NotSmLeft

    bsf     L_DMND
    movlw   d'50'
    call    Delay2Nms
    bcf     L_DMND
    call    CheckLimits
    goto    DoneCommand

NotSmLeft
    movlw   "."             ;
    subwf   RxChar,W          ;"." rotate right for 100ms
    btfss   STATUS, Z
    goto    NotSmRight

    bsf     R_DMND
    movlw   d'50'
    call    Delay2Nms
    bcf     R_DMND
    call    CheckLimits
    goto    DoneCommand

NotSmRight                  
    btfsc   LONGCOMMAND     ;If not a quick command, look for [cr]
    goto    DoLongCmd       ;If not a [cr], do nothing as data has been
                            ;  pushed onto the stack.
    clrf    RxChar          ;Otherwise, clean up, 
    bsf     INTCON, T0IE    ;  re-enable the display multiplex
    goto    MainLoop        ;  and carry on, with no display change
;.............
DoLongCmd
    call    DecodeCommand

DoneCommand
    bcf     NEWCHAR
    bcf     LONGCOMMAND
    clrf    RxChar
    movlw   0xA0            ;Re-position at start of stack
    movwf   StackPointer
ContCmd 
    bsf     INTCON, T0IE    ;Re enable the display multiplex
    goto    DoneSwitches    ;Same as if a switch was pressed
;--------------------------------
StartUp     ;Do all the setting up down here to get it off the front page.
    bcf     STATUS, RP1
    bsf     STATUS, RP0     ;ram page 1
    movlw   b'00001111'     ;4 analogue inputs, A4 spare O/P, A5 Sw Strobe 
    movwf   PORTA       
    movlw   b'11111111'     ;Set as inputs for display polarity detection
    movwf   PORTB           ;   and reading switches
    movlw   b'10000000'     ;C7  RS232 input, 
    movwf   PORTC       

    movlw   b'01000011'     ;Internal input, Prescalar /16,  pull-ups
    movwf   OPTION_REG

    movlw   0x8D            ;AN0,AN1,Ext Refs analogue,
    movwf   ADCON1          ;  right justified result


    movlw   b'00100100'     ;8 bit,TxEN,Async,HiSpeed
    movwf   TXSTA
    movlw   d'25'           ;9600 Baud from 4MHx xtal
    movwf   SPBRG

    bsf     PIE1, RCIE      ;Serial interface received character

    bcf     STATUS , RP0
    clrf    Flags

    movlw   b'10010000'      ;Serial port enabled, continuous receive
    movwf   RCSTA

			;Detect Display polarity
    bsf     SWStrb
    bcf     COMMONANODE
    clrf    PORTC           ;Set all digit drives low

	movlw	d'50'
	call	Delay2Nms

    movf    SEGMENTS, W ;Will pull PORTB low for CC display
    andlw   0x00    ;test polarity
    btfsc   STATUS, Z
    bsf     COMMONANODE
	
			;Preset Display registers
    clrf    IntCounter
	movlw	Digit0
	movwf	FSR
	movlw	d'22'		;t
	call	SegTable
	movwf	INDF

	incf	FSR
	movlw	d'19'		;n
	call	SegTable
	movwf	INDF

	incf	FSR
    movlw   d'17'       ;j
	call	SegTable
	movwf	INDF

	incf	FSR
	movlw	d'25'		;-
	call	SegTable
	movwf	INDF		;Start up display   -jnt

    movlw   0xA0
    movwf   StackPointer    ;To store received serial characters

    movlw   b'10100000'     ;TMR0 Overflow (T0IE) int 
    movwf   INTCON      ; so display starts now

    bsf     SpareD
	movlw	d'250'		;Display intro message for 1s
	call	Delay2Nms
	movlw	d'250'
	call	Delay2Nms

    call    CrLf
    call    CrLf
    movlw   LOW(TextIntro)
    call    DispMsg
    call    CrLf
    call    CrLf
    movlw   LOW(TextConst)
    call    DispMsg
    call    CrLf
    call    SendBackEE

    movlw   LOW(TextHelp)
    call    DispMsg
    call    CrLf
    call    CrLf

    btfss   RCSTA, OERR     ;Check for any overrun error
    goto    StartSerOK      ;If RS232 interface is not connected at sw on.
    bcf     RCSTA, CREN     ;In which case, reset the UART receiver
    nop
    bsf     RCSTA, CREN
StartSerOK

    movlw   GO_LEFT + GO_RIGHT    ;L and Right down at turn on
    subwf   SWStates, W
    btfsc   STATUS, Z
    goto    SetNewZ         ;Allow Z to be changed from Keys


DoneNewZ
    call    ReadP           ;Rotor position in P0/1
    movf    P0, W
    movwf   LastP
    call    CalcBearing     ;Converts P to Bearing (0-3600 in 0.1deg) in A
    call    BinToBCD        ;Binary value in A to BCD variables

    call    ShowBCD4
    movlw   b'01100011'     ;Generate a degree symbol
    movwf   Digit0          ;Blank decimal degrees

    movlw   0xFF            ;To force a display and RS232 update
    movwf   SWStates

    goto    MainLoop
;-----------------------
SetNewZ                     ;Get here if L and R buttons pressed at turn on
    movlw   5
    call    SegTable        ;Display SetZ for 1s
    movwf   Digit3
    movlw   d'14'
    call    SegTable
    movwf   Digit2
    movlw   d'22'
    call    SegTable
    movwf   Digit1
    movlw   2
    call    SegTable
    movwf   Digit0
    movlw   d'250'
    call    Delay2Nms    
    movlw   d'250'
    call    Delay2Nms    


    movlw   Zconst          ;Recover value
    call    GetEE870
    movwf   Q1              ;Work with modified Z value in Q1/0
    movlw   Zconst + 1
    call    GetEE870
    movwf   Q0

SetZloop                    ;Incr / Decr with R/L buttons to change
    movf    Q0, W
    movwf   A0
    movf    Q1, W
    movwf   A1
    call    BinToBCD        ;Use slow button to store
    call    ShowBCD4
    movlw   b'01001000'     ;
    movwf   Digit0          

    movlw   d'10'          ;Incr / Decr 0.1deg values at  20ms intervals
    call    Delay2Nms

    movlw   GO_RIGHT
    subwf   SWStates, W
    btfsc   STATUS, Z
    goto    IncrNewZ

    movlw   GO_LEFT
    subwf   SWStates, W
    btfsc   STATUS, Z
    goto    DecrNewZ

    movlw   GO_SLOW
    subwf   SWStates, W
    btfsc   STATUS, Z
    goto    SaveNewZ

    goto    SetZloop
;..........
DecrNewZ
    decf    Q0
    comf    Q0, W       ;Test for Q0 overflowing from 00 to FF 
    btfsc   STATUS, Z   ;  by  complementing to zero 
    decf    Q1

    comf    Q1, W       ;Now test for going below 0, by looking for Q1 = FF
    btfss   STATUS, Z
    goto    SetZloop

    movlw   LOW(d'3600')    ;In which case set it to maximum
    movwf   Q0
    movlw   HIGH(d'3600')
    movwf   Q1

    goto    SetZloop
;..........
IncrNewZ
    incf    Q0
    btfsc   STATUS, Z
    incf    Q1

    movlw   LOW(d'3600')    ;Test for Q > 3600
    subwf   Q0, W           ;W = Q0 - LOW(d'3600')
    movlw   0               ;If LSByte overflows (C=0), preload W with 1
    btfss   STATUS, C
    movlw   1
    addlw   HIGH(d'3600')   
    subwf   Q1, W           ;W = Q1 - HIGH(d'3600') - Borrow
    btfss   STATUS, C       ;If overflowed (C=0), Q is < 3600, so keep
    goto    SetZloop

    clrf    Q0              ;Otherwise set to zero
    clrf    Q1

    goto    SetZloop
;..........
SaveNewZ
    movlw   Zconst + 1
    movwf   EEAddress
    movf    Q0, W
    call    EEWrite870

    movlw   Zconst
    movwf   EEAddress
    movf    Q1, W
    call    EEWrite870

    call    SendBackEE
    goto    DoneNewZ
;--------------------------------

;===================== End of top level routine ==========================
;=================== Called  Subroutines below here ======================
ReadP
    movlw   b'01000001'      ;AD On, Channel 0
    movwf   ADCON0

    clrf    A0
    clrf    A1
    clrf    A2
    clrf    A3
ReadAD1Loop
    call    HalfBaudDelay   ;Convenient 52us delay for acquisition
    bsf     ADCON0, GO
RdADChan1Loop               
    btfsc   ADCON0, GO
    goto    RdADChan1Loop   ;Loop until Chan 0 conversion is done

    bsf     STATUS, RP0
    movf    ADRESL, W
    bcf     STATUS, RP0
    addwf   A0              ;Add it, in case we later want an average
    btfsc   STATUS, C       ;  of several readings
    incf    A1
    movf    ADRESH, W
    addwf   A1

    movf    A0, W
    movwf   B0              ;Save A/D result in A and B
    movf    A1, W
    movwf   B1

    comf    B1              ;B0/1 = -B0/1
    comf    B0, W
    addlw   1
    movwf   B0
    btfsc   STATUS, Z
    incf    B1              

    movlw   Kconst + 1      ;Low byte of K * 1024
    call    GetEE870        ;  K probably, but not necessarily, = 2
    addwf   B0
    btfsc   STATUS, C
    incf    B1
    movlw   Kconst          ;High byte
    call    GetEE870        ;
    addwf   B1              ;  B  = K * 1024 - B

    movlw   d'12'           ;Multiply by 2^12
    movwf   Counter
ScaleALoop
    bcf     STATUS, C
    rlf     A0
    rlf     A1
    rlf     A2
    rlf     A3
    decfsz  Counter
    goto    ScaleALoop

    call    Divide32x16     ;Q = 2^12 *  N/(2048-N)  (0 - 4095)
                            ;
    movf    Q0, W
    movwf   P0
    movf    Q1, W
    movwf   P1

    return
;---------------------------------------------
CalcBearing            ;From P, gives result in A
    movf    P0, W
    movwf   A0
    movf    P1, W
    movwf   A1

    movlw   LOW(Mconst)     ;Get M = Gradient * 4096 (units of 0.1 degrees)
    call    GetEE870
    movwf   B1              ;High byte of M * 4096
    movlw   LOW(Mconst) + 1
    call    GetEE870
    movwf   B0              ;Low byte of M * 4096
    call    Multiply16      ;M * P * 4096  in  Q0/1/2/3
                            ;  max value 4096 * 3800 = 16M,

    movlw   d'14'           ;Re scale the result 
    movwf   Counter
ScaleQLoop
    bcf     STATUS, C
    rrf     Q3
    rrf     Q2
    rrf     Q1
    rrf     Q0
    decfsz  Counter         ;M * P in Q0/1 only by now
    goto    ScaleQLoop

    movf    Q1, W         
    movwf   A1
    movf    Q0, W
    movwf   A0

    movlw   Msign           ;Get sign of gradient
    call    GetEE870
    iorlw   0
    btfsc   STATUS, Z
    goto    PosGradient

    comf    A0              ;Negate if necessary
    comf    A1
    incf    A0
    btfsc   STATUS, Z
    incf    A1

PosGradient

    movlw   LOW(Cconst) + 1 ;Get LSByte of C, offset value
    call    GetEE870
    addwf   A0
    btfsc   STATUS, C
    incf    A1
    movlw   LOW(Cconst)
    call    GetEE870
    addwf   A1              ;Th, in units of 3600 degrees

    movlw   LOW(Zconst) + 1 ;Get LSByte of bearing Zero value
    call    GetEE870
    addwf   A0
    btfsc   STATUS, C
    incf    A1
    movlw   LOW(Zconst)
    call    GetEE870
    addwf   A1              ;A is correct bearing, possibly +3600 (0.1 deg)

    movf    A0, W           ;Save a copy in B1/0 to work on
    movwf   B0
    movf    A1, W
    movwf   B1

    movlw   0xF0            ; B = B - 3600 (+61936 = 0xF1F0)
    addwf   B0
    btfsc   STATUS, C
    incf    B1
    movlw   0xF1
    addwf   B1              ;If carry is set, then use B else keep A
    btfss   STATUS, C
    goto    NoCorrect360

    movf    B0, W           ;B contains modified bearing
    movwf   A0              ;  
    movf    B1, W
    movwf   A1

NoCorrect360
    return
;-------------------------------------------------------------------------
ShowBCD4                    ;Place contents of BCDMe/Lo on LED display
    bcf     INTCON, T0IE    ;Save flickering while updating
	clrf	DigitPatt
    call    SetDigit        ;Force all digit drives to off state
    
    swapf   BCDMe, W
    andlw   0x0F
    call    SegTable      
    movwf   Digit3          ;MS Digit (4)
ShowBCD3                    ;Just the three LSnibbles on LED display
    movf    BCDMe, W
    andlw   0x0F
    call    SegTable        ;Digit 3
    movwf   Digit2

    swapf   BCDLo, W
    andlw   0x0F
    call    SegTable      
    movwf   Digit1

    movf    BCDLo, W
    andlw   0x0F
    call    SegTable        ;LS Digit (1)
    movwf   Digit0
    bsf     INTCON, T0IE    ;Restore the interrupt
    return
;--------------------------------
SendFullBCD
    movf    BCDHi, W
    andlw   0x0F
    addlw   0x30
    call    Send232
    
SendBCD
    swapf   BCDMe, W
    andlw   0x0F
    addlw   0x30
    call    Send232

    movf    BCDMe, W
    andlw   0x0F
    addlw   0x30
    call    Send232

    swapf   BCDLo, W
    andlw   0x0F
    addlw   0x30
    call    Send232

    movf    BCDLo, W
    andlw   0x0F
    addlw   0x30
    call    Send232

    return
;--------------------------------
CheckLimits                 ;If P0 hasn't changed by more than 2 
    bcf     LIMIT           ; set LIMIT flag and signal
    movf    P0, W           
    subwf   LastP
    btfsc   STATUS, C
    goto    ResultPos

    comf    LastP
    incf    LastP
ResultPos                   ;LastP = 2's complement difference
    btfss   LastP, 7        ;Now generate the absolute value, by
    goto    DiffPos            
    comf    LastP
    incf    LastP           ;LastP contains absolute value of change
DiffPos
    movlw   2
    subwf   LastP, W        ;W = |/\ P| - 2
    btfsc   STATUS, C       ;If +Ve or zero (C=1), all is OK
    goto    NoLimit

    movlw   TextLimit
    call    DispMsg
    call    CrLf
    bsf     LIMIT

    movlw   b'01000000'     ;Display ----
    movwf   Digit0
    movwf   Digit1
    movwf   Digit2
    movwf   Digit3

NoLimit
    movf    P0, W           ;Update with latest reading
    movwf   LastP
    return
;--------------------------------
DecodeCommand               ;Decode and action the command from stack
    movlw   0xA0
    movwf   FSR
    movwf   StackPointer

    movlw   "#"             ;Check to see its a valid command header
    subwf   INDF, W
    btfss   STATUS, Z
    return                  ;And just get out if it isn't

EchoCmdLoop                 ;If it is a [cr], echo back and do action
    movf    StackPointer, W 
    movwf   FSR             
    movf    INDF, W

    movwf   Temp
    movlw   0x0D
    subwf   Temp, W
    btfsc   STATUS, Z
    goto    DoneSendBackCmd
    movf    Temp, W
    call    Send232
    incf    StackPointer

    btfsc   StackPointer, 6 ;Check for overflow to 0xC0 in case of 
    goto    DoneSendBackCmd ;  faulty data, eg. at start up

    goto    EchoCmdLoop
;......
DoneSendBackCmd
    call    CrLf          

    movlw   0xA0
    movwf   StackPointer
    movwf   FSR

    incf    StackPointer        ;Point to the command letter
    movf    StackPointer, W
    movwf   FSR

    movlw   "Z"
    subwf   INDF, W
    btfsc   STATUS, Z
    goto    NewZ

    movlw   "C"
    subwf   INDF, W
    btfsc   STATUS, Z
    goto    NewC

    movlw   "M"
    subwf   INDF, W
    btfsc   STATUS, Z
    goto    NewM

    movlw   "K"
    subwf   INDF, W
    btfsc   STATUS, Z
    goto    NewK

    movlw   TextUnrecCmd
    call    DispMsg
    call    CrLf
    movlw   TextHelp
    call    DispMsg
    call    CrLf

    goto    OutNewCmd
;..........
NewZ
    movlw   TextChange
    call    DispMsg
    movlw   "Z"
    call    Send232
    call    CrLf
    call    CrLf

    movlw   0xA2        ;Point to actual data
    movwf   FSR
    call    GetBCDValue

    movlw   Zconst + 1
    movwf   EEAddress
    movf    A0, W
    call    EEWrite870
    movlw   Zconst
    movwf   EEAddress
    movf    A1, W
    call    EEWrite870
    call    SendBackEE
    goto    OutNewCmd
;..........
NewC
    movlw   TextChange
    call    DispMsg
    movlw   "C"
    call    Send232
    call    CrLf
    call    CrLf

    movlw   0xA2        ;Point to actual data
    movwf   FSR
    call    GetBCDValue

    movlw   Cconst + 1
    movwf   EEAddress
    movf    A0, W
    call    EEWrite870
    movlw   Cconst
    movwf   EEAddress
    movf    A1, W
    call    EEWrite870
    call    SendBackEE
    goto    OutNewCmd
;..........
NewM
    movlw   TextChange
    call    DispMsg
    movlw   "M"
    call    Send232
    call    CrLf
    call    CrLf

    movlw   0xA2        ;Point to actual data
    movwf   FSR
    call    GetBCDValue

    movlw   Mconst + 1
    movwf   EEAddress
    movf    A0, W
    call    EEWrite870
    movlw   Mconst
    movwf   EEAddress
    movf    A1, W
    call    EEWrite870
    movlw   Msign
    movwf   EEAddress
    movlw   0
    btfsc   NEGFLAG
    movlw   1
    call    EEWrite870

    call    SendBackEE
    goto    OutNewCmd
;..........
NewK
    movlw   TextChange
    call    DispMsg
    movlw   "K"
    call    Send232
    call    CrLf
    call    CrLf

    movlw   0xA2        ;Point to actual data
    movwf   FSR
    call    GetBCDValue

    movlw   Kconst + 1
    movwf   EEAddress
    movf    A0, W
    call    EEWrite870
    movlw   Kconst
    movwf   EEAddress
    movf    A1, W
    call    EEWrite870
    call    SendBackEE
    goto    OutNewCmd
;..........
OutNewCmd
    call    CrLf
    movlw   0xA0
    movwf   StackPointer

    return
;--------------------------------
GetBCDValue             ;Extract digits from stack, convert to value
    bcf     NEGFLAG
BcdAlignLoop
    movlw   0x0D
    subwf   INDF, W
    btfsc   STATUS, Z
    goto    BcdTermFound
    incf    FSR
    goto    BcdAlignLoop

BcdTermFound
    decf    FSR         ;Pointing at LS Digit, must have 
    movf    INDF, W     ;  four digits at least
    andlw   0x0F
    movwf   BCDLo

    decf    FSR         
    swapf   INDF, W
    andlw   0xF0
    addwf   BCDLo

    decf    FSR         
    movf    INDF, W
    andlw   0x0F
    movwf   BCDMe

    decf    FSR         
    swapf   INDF, W
    andlw   0xF0
    addwf   BCDMe

    decf    FSR         ;This may be another digit if M, or command

    movlw   "A"
    subwf   INDF, W     ;If >= 'A', then it must be a command letter
    btfsc   STATUS, C   ;W = char - 'A', if +ve (C=1) get out
    goto    DoneBcdValue

    movlw   "0"         ;If <= '0' then it is not a digit, or command so 
    subwf   INDF, W     ;W = char - '0', if -ve (C=0) treat as sign
    btfss   STATUS, C
    goto    WasASign

    movf    INDF, W
    andlw   0x0F
    movwf   BCDHi

    decf    FSR         ;This may be another a sign, or command
WasASign
    movlw   "-"
    subwf   INDF, W
    btfsc   STATUS, Z
    bsf     NEGFLAG
    movlw   "+"
    subwf   INDF, W
    btfsc   STATUS, Z
    bcf     NEGFLAG

DoneBcdValue
    call    BCDToBin

;    bcf     STATUS, C       ;############### DEBUG #############
;    rlf     A0
;    rlf     A1
;    call    BinToBCD
;    movlw   "+"
;    btfsc   NEGFLAG
;    movlw   "-"
;    call    Send232
;    call    SendFullBCD         ;###   for now just loop round the data * 2
;    call    CrLf
;    call    CrLf            ;########### DEBUG ################

    return      ;With decimal value in A1/0 and sign in NEGFLAG
;--------------------------------
Send232
    btfss   PIR1, TXIF      ;Loop until Tx buffer is empty
    goto    Send232  
    movwf   TXREG
    return
;--------------------------------
CrLf
    movlw   0x0D
    call    Send232
    movlw   0x0A
    call    Send232
    return
;--------------------------------
GetEE870   ;New EE routines needed as banks 2/3 used
    bsf     STATUS, RP1     ;Address (0 - 63) in W, result in W
    bcf     STATUS, RP0     ;Bank 2
    movwf   EEADR
    bsf     STATUS, RP0     ;Bank 3
    bcf     EECON1, EEPGD   ;Point to Data memory
    bsf     EECON1, RD      ;EE Read
    bcf     STATUS, RP0     ;Bank2
    movf    EEDATA, w
    bcf     STATUS, RP1     ;Reset to Bank 0
    return
;--------------------------------
EEWrite870	;Writes data in W to address in EEAddress
    bsf     STATUS, RP1 ;
    bcf     STATUS, RP0 ;Bank 2
    movwf   EEDATA      ;
    bcf     STATUS, RP1     ;Bank 0
    movf    EEAddress, w
    bsf     STATUS, RP1     ;Bank2 
    movwf   EEADR
    bsf     STATUS, RP0     ;Bank 3
    bcf     EECON1, EEPGD   ;
    bsf     EECON1, WREN    ;
    bcf     INTCON, GIE

    movlw   0x55
    movwf   EECON2
    movlw   0xAA
    movwf   EECON2
    bsf     EECON1, WR
EEWaitW
    btfsc   EECON1, WR      ;WR is cleared in hardware when write is complete
    goto    EEWaitW
    bcf     EECON1, WREN

    bcf     STATUS, RP0     ;Back to Bank 0
    bcf     STATUS, RP1
    bsf     INTCON, GIE     ;restore Interrupts
	return
;---------------------------

;-------------------------

;--------------------------------
HalfBaudDelay          ;For RS232 9600 Baud link, needs (93) 2* 47 clocks
    movlw   d'12'     ; delay = 3.N + 7 = 43 clocks here, + 8 in routine 
    movwf   DelCount
HalfBaudDelLoop
    decfsz  DelCount
    goto    HalfBaudDelLoop             ;Was 121 for 1200 baud
    nop
    nop
    return
;---------------------------
Delay1ms             ;Cycle time = 1.117us
    movlw   d'178'       ;Delay = 5.N + 5 = 895 clocks = 1000us
    movwf   DelCount     
Delay1msLoop             
    nop
    nop
    decfsz  DelCount
    Goto    Delay1msLoop

    return
;---------------------------
Delay2Nms            ;Callable delay routine.  Delays by 
    movwf   DelCount2    ;  2 * Delcount2 ms (max 511ms)
LoopLongDel              ;For longer delays use timer interrupt
    call    Delay1ms
    call    Delay1ms
    decfsz  DelCount2
    goto    LoopLongDel
    return
;--------------------------
SendBackEE
    movlw   "Z"
    call    Send232
    movlw   " "
    call    Send232
    movlw   Zconst
    call    GetEE870
    movwf   A1
    movlw   Zconst + 1
    call    GetEE870
    movwf   A0
    call    BinToBCD
    call    SendBCD
    call    CrLf

    movlw   "M"
    call    Send232
    movlw   " "
    call    Send232

    movlw   Msign
    call    GetEE870
    iorlw   0
    movlw   "+"
    btfss   STATUS, Z
    movlw   "-"
    call    Send232

    movlw   Mconst
    call    GetEE870
    movwf   A1
    movlw   Mconst + 1
    call    GetEE870
    movwf   A0
    call    BinToBCD
    call    SendFullBCD
    call    CrLf

    movlw   "C"
    call    Send232
    movlw   " "
    call    Send232
    movlw   Cconst
    call    GetEE870
    movwf   A1
    movlw   Cconst + 1
    call    GetEE870
    movwf   A0
    call    BinToBCD
    call    SendBCD
    call    CrLf

    movlw   "K"
    call    Send232
    movlw   " "
    call    Send232
    movlw   Kconst
    call    GetEE870
    movwf   A1
    movlw   Kconst + 1
    call    GetEE870
    movwf   A0
    call    BinToBCD
    call    SendBCD
    call    CrLf
    call    CrLf
    return
;------------------------------------------
BCDToBin    ;Packed BCD in BCDHi/Me/Lo to 16 bit binary in A1/0
    clrf    A1

    movf    BCDHi, W    ;
    andlw   0x0F    ;Mask off 10000s digit
    movwf   A0
    call    Mult10      ;

    swapf   BCDMe, W    ;get 1000s digit
    andlw   0x0F
    addwf   A0     ;Can't overflow at this point as max possible = 99
    call    Mult10

    movf    BCDMe, W    ;100s digit
    andlw   0x0F
    addwf   A0
    btfsc   STATUS, C
    incf    A1
    call    Mult10

    swapf   BCDLo, W    ;10s digit
    andlw   0x0F
    addwf   A0
    btfsc   STATUS, C
    incf    A1
    call    Mult10

    movf    BCDLo, W    ;Units digit
    andlw   0x0F
    addwf   A0
    btfsc   STATUS, C
    incf    A1

    return
;....................
Mult10          ;Multiplies A1/0 by 10, Uses Whi
    movf    A1, W
    movwf   Whi
    movf    A0, W      ;Store original values in Whi/W

    bcf     STATUS, C
    rlf     A0
    rlf     A1
    rlf     A0
    rlf     A1     ;A = 4 * A

    addwf   A0
    btfsc   STATUS, C
    incf    A1
    movf    Whi, W
    addwf   A1     ;A = 5 * A

    bcf     STATUS, C   
    rlf     A0
    rlf     A1     ;A = 10 * A
    return
;--------------------------
;--------------------------------------------
Multiply16    ;16 bit * 16 bit multiplication, more efficient method
    clrf    Q0        
    clrf    Q1        ;A and B 16 bit    
    clrf    Q2
    clrf    Q3
                        ;Up to 328 clock cycles max
    movlw   d'16'       
    movwf   Counter
MultLoop            
    bcf     STATUS , C
    rrf     A1         ;Working bit into carry
    rrf     A0         ;Repeatedly adds shifted version of B
    btfss   STATUS , C  ;  into Q depending on working bit of A
    goto    NoMult

    movf    B0 , W     ;get shifted B0 value
    addwf   Q2        ; and add into accumulator if 1 in 
    btfss   STATUS , C  ; appropriate bit in Avalue
    goto    MulAdd1
    movlw   1
    addwf   Q3
MulAdd1
    movf    B1 , W
    addwf   Q3

NoMult                  ;Need to rotate in carry from final addition
    rrf     Q3        ;C already zero if we've jumped in here
    rrf     Q2
    rrf     Q1
    rrf     Q0
    decfsz  Counter
    goto    MultLoop

    return              ;Q4/3/2/1 = A * B     A kept, B destroyed
;------------------------------------------------------------------

;---------------------------
Divide32x16              ;[32bit] A / [16 bit] B  = [32 bit] Quotient, 
    movlw   d'32'       ; remainder [16 bit] in R
    movwf   Counter
    clrf    R0
    clrf    R1

Divide16Loop
    bcf     STATUS, C
    rlf     A0
    rlf     A1
    rlf     A2
    rlf     A3
    rlf     R0
    rlf     R1              ;Rotate A through Remainder

    movf    R0, W           ;Make backup of original R0/1
    movwf   Temp
    movf    R1, W
    movwf   Temp2


    movf    B0, W           
    subwf   R0
    btfsc   STATUS, C
    goto    NotDiv16Ovf
    decf    R1

    comf    R1, W       ;if borrowed, R1 will go to 0xFF
    btfsc   STATUS, Z   ; so test for 0xFF by complementing it to zero (in W)
    goto    RestoreR

NotDiv16Ovf                 ;R0/1 contains the result of the subtraction 
    movf    B1, W           ;  irrespective of any overflow or not.
    subwf   R1              

    btfsc   STATUS, C
    goto    DivdPassThru    ;With C=1, result was sucessful

RestoreR
    movf    Temp, W         ;If overflow, have to recover original R1/0
    movwf   R0              ;  from the backup
    movf    Temp2, W        
    movwf   R1              ;Otherwise, leave R as it stands
    bcf     STATUS, C       ;C=0 to be shifted into Q

DivdPassThru
    rlf     Q0              ;Use the carry bit from the subtraction
    rlf     Q1              ; to build up Q
    rlf     Q2
    rlf     Q3
    decfsz  Counter
    goto    Divide16Loop

    return
;------------------------------------
BinToBCD        ;Takes in A1/A0, calculates packed BCD equivalent
    clrf    BCDHi      
    clrf    BCDMe
    clrf    BCDLo
    movlw   d'16'
    movwf   Counter
BcdLoop
    rlf     A0
    rlf     A1
    rlf     BCDLo
    rlf     BCDMe
    rlf     BCDHi
    decfsz  Counter     ;Need a final shift only, so loop count here
    goto    BCDgo
    goto    BCDone
    ;..........
BCDgo
    movf    BCDHi , W
    addlw   3
    movwf   Temp
    btfsc   Temp , 3
    movwf   BCDHi       ;If BCDHI >= 5, add 3 and store back

    movf    BCDHi , W
    addlw   0x30
    movwf   Temp
    btfsc   Temp , 7
    movwf   BCDHi       ;Adjust BCDHI if > 80

    movf    BCDMe , W   ;Repeating for middle then low digits
    addlw   3
    movwf   Temp
    btfsc   Temp , 3
    movwf   BCDMe      

    movf    BCDMe , W
    addlw   0x30
    movwf   Temp
    btfsc   Temp , 7
    movwf   BCDMe     

    movf    BCDLo , W   ;
    addlw   3
    movwf   Temp
    btfsc   Temp , 3
    movwf   BCDLo    

    movf    BCDLo , W
    addlw   0x30
    movwf   Temp
    btfsc   Temp , 7
    movwf   BCDLo   

    goto    BcdLoop
    ;..................
BCDone
    return
;------------------------------------------
;===========================================

;Generating a message from data stored in programme memory

 org 0x6F0    ;Ensure  MsgTable  onwards is all in the same segment

DispMsg   ;Enter with address of Text in W  eg.  'MOVLW LOW(TextLabel)'
    movwf   AddrHold    ;Store offset of text string

    movlw   high(MsgTable)
    movwf   PCLATH

    movlw   low(MsgTable)
    addlw   1
    subwf   AddrHold , W    ;W = AddrTextString - TableStart - 1
                            ;  = Offset of string start in table
MsgLoop
    movwf   MsgCount
    call    MsgTable
    andlw   0xFF            ;end of message
    btfsc   STATUS , Z
    goto    MsgDone

    call    Send232
    incf    MsgCount, W
    goto    MsgLoop
;...........
MsgDone
    return
;.........
 org    0x700
MsgTable
    addwf   PCL
TextIntro       dt "G4JNT Rotator Controller", 0
TextConst       dt "Stored Constants -" , 0
TextChange      dt "Change ", 0
TextUnrecCmd    dt "Unrecognised command...",0
TextHelp        dt "Commands # + Z/C/K/M + (-)nnnn(n)  ", 0
TextLimit       dt "Limit",0
;-------------------------------    

;--------------------------------
SegTable      ;Call with digit / letter / symbol code, return with 
    andlw   0x1F    ;  segment pattern
	movwf	Temp
	movlw	HIGH($)
	movwf	PCLATH
	
	movf    Temp , W
    addwf   PCL
	
        ;  gfedcba  ;    aaaa
    dt  b'00111111' ;0  f    b
    dt  b'00000110' ;1  f    b
    dt  b'01011011' ;2  f    b
    dt  b'01001111' ;3  f    b
    dt  b'01100110' ;4   gggg
    dt  b'01101101' ;5  e    c  
    dt  b'01111101' ;6  e    c 
    dt  b'00000111' ;7  e    c
    dt  b'01111111' ;8  e    c 
    dt  b'01101111' ;9   dddd 
	dt	b'01110111'	;A	10	 			
	dt	b'01111100'	;b
	dt	b'00111001'	;c
	dt	b'01011110'	;d
	dt	b'01111001'	;E
	dt	b'01111000'	;F	15
	dt	b'01110110'	;H
	dt	b'00001110'	;j
	dt	b'00111000'	;L
	dt	b'01010100'	;n
	dt	b'01110011'	;P	20
	dt	b'01010000'	;r
	dt	b'01111000'	;t
	dt	b'00111110'	;U
	dt	b'01100110'	;y
	dt	b'01000000'	;-	25
	dt	b'01010000'	;=
	dt	b'00000001'	;[top]
	dt	b'00000000'	;[bottom]
	dt	b'01000000'	;[middle]
	dt	b'00000000'	;[off]	30
	dt	b'00000000'	;[off]

MsgTableEnd

    if ((MsgTable & 0xFF) >= (MsgTableEnd & 0xFF))
        ERROR    "Table code crosses page boundary"
        ERROR    "ADJUST START OF TABLE"
    endif

;============================
    cblock      0x20
    Temp
    Temp2
    Whi
    StsReg
    Wreg
    FSRReg
    SegmentPatt
    DigitPatt
    IntCounter
	SWStates
    SWStatesOld
    DelCount
    DelCount2
	Ccache
    Counter         ;general purpose
    Counter2
    MsgCount
    AddrHold
    Flags
    RxChar
    EEAddress
    EEContents
    StackPointer    ;Serial characters 
    BearLo
    BearHi
    A0
    A1
    A2
    A3
    B0
    B1
    X0
    X1
    R0
    R1
    Q0
    Q1
    Q2
    Q3
    P0              ;Pot. rotor position  0 - (1023)
    P1
    LastP       ;Used in limit checking
    BCDLo
    BCDMe
    BCDHi

	Digit0		;Four display digits, accessed indirectly in mpx routine
	Digit1		;So must be consecutive
	Digit2
	Digit3
    endc

    end
